/*
 * lpcload.c - load firmware into ram of lpc2220 via uart0
 *
 * author: hackbard@hackdaworld.org, rolf.anders@physik.uni-augsburg.de
 *
 * build: make
 * usage: sudo ./lpcload -d /dev/ttyS0 -f firmware.hex -D0 fw0.bin [-v]
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>

#define VERBOSE			(1<<0)
#define FIRMWARE		(1<<1)
#define BANK0			(1<<2)
#define BANK2			(1<<3)
#define BL			(1<<4)
#define FLASHFW			(1<<5)

#define BANK0_ADDR		0x80000000
#define BANK2_ADDR		0x82000000
#define BANK_SIZE		0x00100000
#define BL_ADDR			0x7fffe000
#define BL_SIZE			0x00002000

#define CMD_READ		'R'		// stay compatible to fwflash!
#define CMD_WRITE		'W'
#define CMD_CHIP_ERASE		'E'

#define TXRX_TYPE_SYNC		0x00
#define TXRX_TYPE_CKSM		0x00
#define TXRX_TYPE_BAUD		0x01
#define TXRX_TYPE_CMD		0x02
#define TXRX_TYPE_DATA		0x03
#define TXRX_TYPE_GO		0x04

#define CMD_SUCCESS		"0\r\n"
#define INVALID_COMMAND		"1\r\n"
#define SRC_ADDR_ERROR		"2\r\n"
#define DST_ADDR_ERROR		"3\r\n"
#define SRC_ADDR_NOT_MAPPED	"4\r\n"
#define DST_ADDR_NOT_MAPPED	"5\r\n"
#define COUNT_ERROR		"6\r\n"
#define COMPARE_ERROR		"10\r\n"
#define BUSY			"11\r\n"
#define PARAM_ERROR		"12\r\n"
#define ADDR_ERROR		"13\r\n"
#define ADDR_NOT_MAPPED		"14\r\n"
#define CMD_LOCKED		"15\r\n"
#define INVALID_CODE		"16\r\n"
#define INVALID_BAUD_RATE	"17\r\n"
#define INVALID_STOP_BIT	"18\r\n"

#define RAM			0x01
#define FLASH			0x02

#define CRYSTFREQ		"10000"
#define RAMOFFSET		0x40000200

#define BUFSIZE			128

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;

typedef struct s_lpc {
	int sfd;		/* serial fd */
	char sdev[128];		/* seriel device */
	int fwfd;		/* fimrware fd */
	char fwfile[128];	/* firmware file */
	u32 info;		/* info/mode */
	char freq[8];		/* frequency */
	char bank0[127];	/* flash dump bank0 */
	int b0fd;		/* dumpfile fd bank0 */
	char bank2[127];	/* flash dump bank2 */
	int b2fd;		/* dumpfile fd bank0 */
	char bl[127];		/* flash dump bootloader */
	int blfd;		/* dumpfile fd bootloader */
	char ffwfile[127];	/* file with firmware to be flashed */
	int ffwfd;		/* fd of the above */
	u32 roff;		/* ram offset of uc */
	u32 jaddr;		/* addr for the jump */
} t_lpc;

void usage(void) {

	printf("possible argv:\n");
	printf("  -d  <serial device>\n");
	printf("  -f  <firmware>\n");
	printf("  -c  <crystal freq>\n");
	printf("  -Dx <filename>\n");
	printf("      x=0: bank0, x=2: bank2, x=b: bootloader\n");
	printf("  -w <firmware which goes to flash>\n");
	printf("  -v\n");

}

int open_serial_device(t_lpc *lpc) {

	struct termios term;

	//memset(&term,0,sizeof(struct termios));

	/* open serial device */

	lpc->sfd=open(lpc->sdev,O_RDWR);
	if(lpc->sfd<0) {
		perror("tts open");
		return lpc->sfd;
	}

	/* configure the serial device */

	tcgetattr(lpc->sfd,&term);

	// input/output baudrate

	cfsetispeed(&term,B38400);
	cfsetospeed(&term,B38400);

	// control options -> 8n1

	term.c_cflag&=~PARENB;	// no parity
	term.c_cflag&=~CSTOPB;	// only 1 stop bit
	term.c_cflag&=~CSIZE;	// no bit mask for data bits
	term.c_cflag|=CS8;	// 8 data bits

	// line options -> raw input
	
	term.c_lflag&=~(ICANON|ECHO|ECHOE|ISIG);

	// input options -> enable flow control
	
	term.c_iflag&=~(INLCR|ICRNL|IXANY);
	term.c_iflag|=(IXON|IXOFF);
	
	// output options

	term.c_oflag=0;

	// more control options -> timeout / flow control
	
	term.c_cc[VMIN]=0;
	term.c_cc[VTIME]=20;	// 2 seconds timeout
	//term.c_cc[VSTART]=0x11;
	//term.c_cc[VSTOP]=0x13;

	tcsetattr(lpc->sfd,TCSANOW,&term);

	return lpc->sfd;
}

int reconfig_serial_device(t_lpc *lpc) {

	struct termios term;
	int ret;

	/* reconfigure the serial device for our lousy loader tool */

	tcgetattr(lpc->sfd,&term);

	// disable flow control
	
	term.c_iflag&=~(IXON|IXOFF|IXANY|INLCR|ICRNL);

	// change baudrate

	cfsetispeed(&term,B115200);
	cfsetospeed(&term,B115200);

	ret=tcsetattr(lpc->sfd,TCSANOW,&term);

	return ret;
}

int open_firmware(t_lpc *lpc) {

	/* open firmware file */

	lpc->fwfd=open(lpc->fwfile,O_RDONLY);

	if(lpc->fwfd<0) {
		perror("fw open");
		return lpc->fwfd;
	}

	if(lpc->info&FLASHFW) {
		lpc->ffwfd=open(lpc->ffwfile,O_RDONLY);
		if(lpc->ffwfd<0) {
			perror("ffw open");
			return lpc->ffwfd;
		}
	}

	return lpc->fwfd;
}

int open_dumpfiles(t_lpc *lpc) {

	/* open dumpfiles */

	if(lpc->info&BANK0) {
		lpc->b0fd=open(lpc->bank0,O_WRONLY|O_CREAT);
		if(lpc->b0fd<0) {
			perror("bank0 dump file open");
			return lpc->b0fd;
		}
	}

	if(lpc->info&BANK2) {
		lpc->b2fd=open(lpc->bank2,O_WRONLY|O_CREAT);
		if(lpc->b2fd<0) {
			perror("bank2 dump file open");
			return lpc->b2fd;
		}
	}

	if(lpc->info&BL) {
		lpc->blfd=open(lpc->bl,O_WRONLY|O_CREAT);
		if(lpc->blfd<0) {
			perror("bootloader dump file open");
			return lpc->blfd;
		}
	}

	return 0;

}
int txrx(t_lpc *lpc,char *buf,int len,u8 type) {

	int ret,cnt;
	int i;

	/* write */

	if(lpc->info&VERBOSE)
		printf("  >> ");
	cnt=0;
	while(len) {
		ret=write(lpc->sfd,buf+cnt,len);
		if(ret<0) {
			perror("txrx write");
			return ret;
		}
		if(lpc->info&VERBOSE)
			for(i=0;i<ret;i++)
				printf("%c",
				       ((buf[cnt+i]>0x19)&(buf[cnt+i]<0x7f))?
				       buf[cnt+i]:'.');
		len-=ret;
		cnt+=ret;
	}
	if(lpc->info&VERBOSE) {
		printf(" | ");
		for(i=0;i<cnt;i++)
			printf("%02x ",buf[i]);
		printf("| (%d)\n",cnt);
	}



	/* cut the echo if not of type auto baud */

	if(type!=TXRX_TYPE_BAUD) {
		while(cnt) {
			ret=read(lpc->sfd,buf,cnt);
			if(ret<0) {
				perror("txrx echo cut");
				return ret;
			}
			cnt-=ret;
		}
	}

	/* return if type is go */

	if(type==TXRX_TYPE_GO)
		return cnt;

	/* return here if type is data */

	if(type==TXRX_TYPE_DATA)
		return cnt;

	/* read */

	ret=read(lpc->sfd,buf,1);
	if(ret<0) {
		perror("txrx read (first byte)");
		return ret;
	}
		
	switch(buf[0]) {
		case 'S':
			cnt=13;
			break;
		case 'O':
			cnt=3;
			break;
		case 'R':
			cnt=7;
			break;
		case '0':
			cnt=2;
			break;
		default:
			printf("txrx read: bad return byte '%02x'\n",buf[0]);
			break;
	}

	ret=1;
	i=cnt;
	while(i) {
		ret=read(lpc->sfd,buf+1+cnt-i,i);
		if(ret<0) {
			perror("txrx read (next bytes)");
			return ret;
		}
		i-=ret;
	}
	if(lpc->info&VERBOSE) {
		printf("  << ");
		for(i=0;i<cnt+1;i++)
			printf("%c",((buf[i]>0x19)&(buf[i]<0x7f))?
			            buf[i]:'.');
		printf(" | ");
		for(i=0;i<cnt+1;i++)
			printf("%02x ",buf[i]);
		printf("| (%d)\n",cnt+1);
	}
	buf[cnt+1]='\0';

	/* check/strip return code if type is cmd */

	if(type==TXRX_TYPE_CMD) {
		ret=strlen(CMD_SUCCESS);
		if(!strncmp(buf,CMD_SUCCESS,ret)) {
			for(i=ret;i<cnt;i++)
				buf[i-ret]=buf[i];
			buf[cnt]='\0';
		}
		else {
			printf("txrx bad return code!\n");
			return -1;
		}
	}

	return cnt;
}

int bl_init(t_lpc *lpc) {

	char buf[BUFSIZE];
	int len;

	/* auto baud sequence */
	buf[0]='?';
	txrx(lpc,buf,1,TXRX_TYPE_BAUD);
	if(strncmp(buf,"Synchronized\r\n",14)) {
		printf("auto baud detection failed\n");
		return -1;
	}

	/* tell bl that we are synchronized (it's allready in buf) */
	txrx(lpc,buf,14,TXRX_TYPE_SYNC);
	if(strncmp(buf,"OK\r\n",4)) {
		printf("sync failed\n");
		return -1;
	}

	/* tell bl the crystal frequency */
	len=strlen(lpc->freq)+2;
	strncpy(buf,lpc->freq,BUFSIZE);
	buf[len-2]='\r';
	buf[len-1]='\n';
	txrx(lpc,buf,len,TXRX_TYPE_SYNC);
	if(strncmp(buf,"OK\r\n",4)) {
		printf("freq set failed\n");
		return -1;
	}

	return 0;
}

int unlock_go(t_lpc *lpc) {

	char buf[BUFSIZE];
	int ret;

	memcpy(buf,"U 23130\r\n",9);
	ret=txrx(lpc,buf,9,TXRX_TYPE_CMD);

	return ret;
}

int go(t_lpc *lpc) {

	char buf[BUFSIZE];
	int ret,len;

	snprintf(buf,BUFSIZE,"G %d A\r\n",lpc->jaddr);
	len=strlen(buf);
	ret=txrx(lpc,buf,len,TXRX_TYPE_GO);

	return ret;
}

int uuencode(u8 *in,u8 *out,int len) {

	out[0]=0x20+len;
	out[1]=0x20+((in[0]>>2)&0x3f);
	out[2]=0x20+(((in[0]<<4)|(in[1]>>4))&0x3f);
	out[3]=0x20+(((in[1]<<2)|(in[2]>>6))&0x3f);
	out[4]=0x20+(in[2]&0x3f);

	return 0;
}

int write_to_ram(t_lpc *lpc,char *buf,u32 addr,int len) {

	int lcount;
	u32 checksum;
	char txrxbuf[BUFSIZE];
	int count,bcnt;
	int nlen,slen;
	int i;

	/* check length */
	if(len%4) {
		printf("ram write: not a multiple of 4\n");
		return -1;
	}

	/* make it a multiple of 3 (reason: uuencode) */
	nlen=(!(len%3))?len:((len/3+1)*3);
	if(nlen>BUFSIZE) {
		printf("ram write: too much data\n");
		return -1;
	}
	for(i=len;i<nlen;i++) buf[i]=0;

	/* prepare addr */
	addr+=lpc->roff;

	/* prepare write command */
	if(lpc->info&VERBOSE)
		printf("writing 0x%02x bytes to 0x%08x\n",len,addr);
	snprintf(txrxbuf,BUFSIZE,"W %d %d\r\n",addr,len);
	slen=strlen(txrxbuf);

	/* send command and check return code */
	txrx(lpc,txrxbuf,slen,TXRX_TYPE_CMD);

	/* send data */
	lcount=0;
	bcnt=0;
	count=0;
	checksum=0;
	while(bcnt<nlen) {

		/* uuencode / prepare data bytes */
		uuencode((u8 *)(buf+bcnt),(u8 *)(txrxbuf),
		         (bcnt==nlen-3)?(len%3?len%3:3):3);
		txrxbuf[5]='\r';
		txrxbuf[6]='\n';

		/* checksum */
		checksum+=((u8)buf[bcnt]+(u8)buf[bcnt+1]+(u8)buf[bcnt+2]);

		/* send a data line */
		txrx(lpc,txrxbuf,7,TXRX_TYPE_DATA);

		/* increase counters */
		lcount+=1;
		bcnt+=3;
		count+=3;

		/* checksum */
		if((!(lcount%20))|(bcnt==nlen)) {
			/* send backtick */
			memcpy(txrxbuf,"`\r\n",3);
			//txrx(lpc,txrxbuf,3,TXRX_TYPE_DATA);
			/* send checksum */
			snprintf(txrxbuf,BUFSIZE,"%d\r\n",checksum);
			slen=strlen(txrxbuf);
			txrx(lpc,txrxbuf,slen,TXRX_TYPE_CKSM);
			if(!strncmp(txrxbuf,"RESE",4)) {
				read(lpc->sfd,txrxbuf+4,4);
				printf("ram write: resending ...\n");
				bcnt-=count;
			}
			if(strncmp(txrxbuf,"OK\r\n",4)) {
				printf("ram write: bad response\n");
				return -1;
			}
			/* reset checksum & counter */
			checksum=0;
			count=0;
		}

	}

	return 0;
}

int send_cmd(int sfd,u32 addr,u32 len,u8 cmd) {

	int ret,i,cnt,size;
	int as,ls;
	u8 send[9];

	switch(cmd) {
		case CMD_READ:
		case CMD_WRITE:
			as=4;
			ls=4;
			break;
		case CMD_CHIP_ERASE:
			as=0;
			ls=1;
			break;
		default:
			printf("send cmd: cmd '%02x' not supported\n",cmd);
			return -1;
	}

	size=1+as+ls;
	send[0]=cmd;

	for(i=0;i<as;i++)
		send[1+i]=(addr>>((as-1-i)*8))&0xff;
	for(i=0;i<ls;i++)
		send[1+i+as]=(len>>((ls-1-i)*8))&0xff;

	cnt=0;
	printf("  sending cmd: ");
	while(size) {
		ret=write(sfd,send+cnt,size);
		if(ret<0) {
			perror("dump file: send cmd ");
			return ret;
		}
		for(i=cnt;i<cnt+ret;i++)
			printf("%02x ",send[i]);
		size-=ret;
		cnt+=ret;
	}
	printf("\n");

	return 0;
}

int write_to_flash(t_lpc *lpc,char *buf,u32 addr,int len) {

	int cnt,size,ret;

	/* send cmd */
	send_cmd(lpc->sfd,addr,len,CMD_WRITE);

	/* transfer data */
	while(len) {
		size=2;
		while(size) {
			ret=write(lpc->sfd,buf+cnt,2);
			if(ret<0) {
				perror("transmit flash content");
				return ret;
			}
			size-=ret;
		}
		cnt+=2;
	}

	return 0;
}

int firmware_to_mem(t_lpc *lpc,u8 memtype) {

	char buf[BUFSIZE];
	u32 addr,len,type;
	int ret,temp;
	int fd;

	/* prepare for memory type */
	if(memtype==RAM)
		fd=lpc->fwfd;
	else if(memtype==FLASH)
		fd=lpc->ffwfd;
	else
		return -1;

	/* read a line */
	ret=1;
	while(ret) {
		/* sync line */
		ret=read(fd,buf,1);
		switch(buf[0]) {
			case '\r':
				continue;
			case '\n':
				continue;
			case ':':
				/* start code */
				break;
			default:
				printf("fw to mem: no ihex format\n");
				return -1;
		}
		/* read len */
		ret=read(fd,buf,2);
		sscanf(buf,"%02x",&len);
		/* read addr */
		ret=read(fd,buf,4);
		sscanf(buf,"%04x",&addr);
		/* read type */
		ret=read(fd,buf,2);
		sscanf(buf,"%02x",&type);
		/* successfull return if type is end of file */
		if(type==0x01)
			return 0;
		/* read data (and cksum) */
		ret=read(fd,buf,2*(len+1));
		if(ret!=(2*(len+1))) {
			printf("fw to mem: data missing\n");
				return -1;
		}
		for(ret=0;ret<len;ret++) {
			sscanf(buf+2*ret,"%02x",&temp);
			buf[ret]=temp;
		}
		/* act according to type */
		switch(type) {
			//case 0x03:
			//	/* get cs and ip */
			//	break;
			case 0x00:
				if(len%4) {
					printf("fw to mem: invalid len\n");
					return -1;
				}
				if(memtype==RAM)
					write_to_ram(lpc,buf,addr,len);
				else
					write_to_flash(lpc,buf,addr,len);
				break;
			case 0x04:
				lpc->roff=((buf[0]<<24)|(buf[1]<<16));
				break;
			case 0x05:
				lpc->jaddr=((buf[0]<<24)|(buf[1]<<16));
				lpc->jaddr|=((buf[2]<<8)|buf[3]);
				break;
			default:
				printf("fw to mem: unknown type %02x\n",type);
				return -1;
		}
	}

	return 0;
}

int lpc_txbuf_flush(t_lpc *lpc) {

	int i,ret;
	u8 buf[16];

	ret=1;
	printf("flushing lpc tx buffer: ");
	while(ret) {
		ret=read(lpc->sfd,buf,16);
		for(i=0;i<ret;i++)
			printf("%02x ",buf[i]);
	}
	printf("\n");

	return 0;
}

int dump_files(int sfd,int dfd,u32 addr,u32 len) {

	int ret;
	int size;
	int cnt;
	u8 buf[16];

	printf("dumping content (addr=0x%08x, len=0x%08x) ...\n",addr,len);

	/* send the cmd */
	send_cmd(sfd,addr,len,CMD_READ);

	/* receive data and dump it to file */
	ret=1;
	cnt=0;
	printf("  receiving data ...\n");
	while(ret) {
		ret=read(sfd,buf,16);
		if(ret<0) {
			perror("dump file: read data");
			return ret;
		}
		size=ret;
		cnt=0;
		while(size) {
			ret=write(dfd,buf+cnt,size-cnt);
			if(ret<0) {
				perror("dump file: write data");
				return ret;
			}
			cnt+=ret;
			size-=ret;
		}
	}

	return 0;
}

int main(int argc,char **argv) {

	t_lpc lpc;
	int i;
	int ret;

	/*
	 * initial ... 
	 */

	memset(&lpc,0,sizeof(t_lpc));
	strncpy(lpc.freq,CRYSTFREQ,7);
	lpc.roff=RAMOFFSET;
	lpc.jaddr=RAMOFFSET;

	/* parse argv */

	for(i=1;i<argc;i++) {

		if(argv[i][0]!='-') {
			usage();
			return -1;
		}

		switch(argv[i][1]) {
			case 'd':
				strncpy(lpc.sdev,argv[++i],127);
				break;
			case 'f':
				strncpy(lpc.fwfile,argv[++i],127);
				lpc.info|=FIRMWARE;
				break;
			case 'v':
				lpc.info|=VERBOSE;
				break;
			case 'c':
				strncpy(lpc.freq,argv[++i],7);
				break;
			case 'D':
				if(argv[i][2]=='0') {
					lpc.info|=BANK0;
					strncpy(lpc.bank0,argv[++i],127);
					break;
				}
				else if(argv[i][2]=='2') {
					lpc.info|=BANK2;
					strncpy(lpc.bank2,argv[++i],127);
					break;
				}
				else if(argv[i][2]=='b') {
					lpc.info|=BL;
					strncpy(lpc.bl,argv[++i],127);
					break;
				}
				else {
					usage();
					return -1;
				}
			case 'w':
				strncpy(lpc.ffwfile,argv[++i],127);
				lpc.info|=FLASHFW;
				break;
			default:
				usage();
				return -1;
		}

	}

	/* open serial port */
	if(open_serial_device(&lpc)<0)
		goto end;

	/* boot loader init */
	printf("boot loader init ...\n");
	if(bl_init(&lpc)<0)
		return -1;

	/* quit if there is no hex file to process */
	if(!(lpc.info&FIRMWARE)) {
		printf("no firmware -> aborting\n");
		goto end;
	}

	/* open firmware file */
	if(open_firmware(&lpc)<0)
		goto end;

	/* open dump files */
	if(open_dumpfiles(&lpc)<0)
		goto end;

	/* parse intel hex file and write to ram */
	printf("write firmware to ram ...\n");
	firmware_to_mem(&lpc,RAM);

	/* unlock go cmd */
	printf("unlock go command ...\n");
	unlock_go(&lpc);

	/* go! */
	printf("go ...\n");
	ret=go(&lpc);

	/* flush the lpc2220 tx buf */
	lpc_txbuf_flush(&lpc);

	/* reconfigure the serial port */
	if(reconfig_serial_device(&lpc)<0)
		goto end;

	/* download flash/bootloader content */
	if(lpc.info&BANK0)
		dump_files(lpc.sfd,lpc.b0fd,BANK0_ADDR,BANK_SIZE);
	if(lpc.info&BANK2)
		dump_files(lpc.sfd,lpc.b2fd,BANK2_ADDR,BANK_SIZE);
	if(lpc.info&BL)
		dump_files(lpc.sfd,lpc.blfd,BL_ADDR,BL_SIZE);

	/* write a firmware to the lpc flash */
	if(lpc.info&FLASHFW) {
		send_cmd(lpc.sfd,0,0,CMD_CHIP_ERASE);
		lpc.roff=BANK0_ADDR;
		firmware_to_mem(&lpc,FLASH);
	}

end:
	if(lpc.sfd)
		close(lpc.sfd);
	if(lpc.fwfd)
		close(lpc.fwfd);
	if(lpc.b0fd)
		close(lpc.b0fd);
	if(lpc.b2fd)
		close(lpc.b2fd);
	if(lpc.blfd)
		close(lpc.blfd);
	if(lpc.ffwfd)
		close(lpc.ffwfd);

	return 0;
}

